import logging
import warnings
import numpy as np
import hsbalance.tools as tools
import pandas as pd

logger = logging.getLogger(__name__)
logger.propagate = False
logger.setLevel(logging.DEBUG)
console_handle = logging.StreamHandler()
console_handle.setLevel(logging.INFO)
logger.addHandler(console_handle)

pd.set_option('display.max_columns', 1000)  # Set maximum number of columns to 1000
class Alpha():

    """
    Docstring for ALPHA.
    Alpha is the an influence coefficient matrix
    Influence coefficient matrix is a representation of the change of vibration
    vector in a measuring point when putting a unit weight on a balancing plane.
    """

    def __init__(self:'Influence matrix', name:'string'=''):
        """
        Instantiate an instance of Alpha
        name: optional name of Alpha
        """
        self.name = name
        self.value = None

    def add(self, direct_matrix:'np.array'=None, A:'initial_vibration numpy.array'=None,
            B:'trial matrix numpy.array'=None, U:'trial weight row vector numpy.array'=None,
            keep_trial:'optional keep the previous trial weight in every succeeding trial'=False,
            name:'string'=''):
        '''
        Method to add new values for Alpha instance
        either the direct_matrix is needed or ALL of (A, B, U)
        Args:
            direct_matrix: numpy array M rows -> measuring points,
                        N columns -> balancing planes
            A: Initial vibration column array -> numpy array
            B: Trial matrix MxN array -> numpy array
        '''
        self.A = A
        self.B = B
        self.U = U
        self.keep_trial = keep_trial
        try:  # test if direct input
            _ = direct_matrix.shape
            if direct_matrix.ndim < 2:
                raise IndexError('Influence coefficient matrix should be of more than 1 dimensions.')
            if direct_matrix.shape[0] >= direct_matrix.shape[1]:
                self.value = direct_matrix
            else:
                raise tools.CustomError('Number of rows(measuring points) should be '
                                  'equal or  more than the number of columns '
                                  '(balancing planes)!')
            if self.A is not None or self.B is not None or self.U is not None:
                raise ValueError('Either (direct Matrix) or (A, B, U) should be input, but not both.')


        except AttributeError:
            # if direct matrix is not input calculate it from A, B, U
            # test the exstiance of A, A0, B, U to calculate ALPHA
            try:
                all([A.shape, B.shape, U.shape])
                # Test dimensions
                if A.shape[1] > 1:
                    raise tools.CustomError('`A` should be column vector')
                elif U.ndim > 1:
                    raise tools.CustomError('`U` should be row vector')
                elif B.shape[0] != A.shape[0] or B.shape[1] != U.shape[0]:
                    raise tools.CustomError('`B` dimensions should match `A`and `U`')
                else:
                    self.A = A
                    self.B = B
                    self.U = U
                    if not keep_trial:
                        self.value = (self.B - self.A) / self.U
                    else:
                        _A_keep_trial = np.delete((np.insert(self.B, [0], self.A, axis=1)),
                                                 -1, axis=1)
                        self.value = (self.B - _A_keep_trial) / self.U
            except AttributeError:
                raise tools.CustomError('Either direct_matrix or (A,B,U) '
                                        'should be passed "numpy arrays"')
        if self.value is not None:
            self.M = self.value.shape[0]
            self.N = self.value.shape[1]


    def check(self, ill_condition_remove=False):
        '''
        Method to check the alpha value
            * check the symmetrical of the matrix (check for square matrix only,
            for square matrix it should be symmetric obeying the reciprocity law)
            * check for ill conditioned planes:
                if for any reason two or more planes has independent readings
                for example [[1, 2 , 3], [2, 4, 6]] this is named as ill-conditioned planes
                as they does not carry new information from the system and considering them
                cause solution infiltration.
            ill_condition_remove = True : remove the ill_condition planes after the check
        '''
        if self.M == self.N:
            _check_sym = np.allclose(self.value, self.value.T, 0.1, 1e-06)
            if not _check_sym:
                warnings.warn('\nWarning: Influence Matrix is asymmetrical!')
                logger.info('\nInfluence Matrix is asymmetrical, check your data.')
            else:
                logger.info('\nInfluence Matrix is symmetric --> OK')
        else:
            logger.info('\nNot a square matrix --> no exact solution.')

        # Checking ILL-CONDITIONED planes
        ill_plane = tools.ill_condition(self.value)
        if ill_plane:
            logger.info(f'\nIll-conditioned found in plane # {ill_plane}')
            if ill_condition_remove:
                logger.warn(f'\nRemoving Ill-conditioned plane # {ill_plane}')
                logger.info(f'\nIC matrix before removing\n{tools.convert_cart_math(self.value)}\n')
                self.value = np.delete(self.value,[ill_plane], axis=1)
                logger.info(f'\nIC matrix after removing\n{tools.convert_cart_math(self.value)}\n')
        else:
            logger.info('\nNo ill conditioned planes --> ok')

    def _info(self):
        '''
        Method to summarize the results for alpha.
        return generator of tuples(title:str, item)
        '''
        if self.name:
            yield ('Name', self.name)
        if self.value is not None:
            _index = (f'Sensor {m+1}' for m in range(self.value.shape[0]))
            _columns = (f'Plane {n+1}' for n in range(self.value.shape[1]))
            yield ('Coefficient Values', pd.DataFrame(tools.convert_cart_math(self.value),
                        index=_index, columns=_columns))
        if self.A is not None:
            _index = (f'Sensor {m+1}' for m in range(self.A.shape[0]))
            yield ('Initial Vibration', pd.DataFrame(tools.convert_cart_math(self.A),
                        index=_index, columns=['Vibration']))
        if self.B is not None:
            _index = (f'Sensor {m+1}' for m in range(self.B.shape[0]))
            _columns = (f'Plane {n+1}' for n in range(self.B.shape[1]))
            yield ('Trial Runs Vibration', pd.DataFrame(tools.convert_cart_math(self.B),
                        index=_index, columns=_columns))
        if self.U is not None:
            _index = (f'Plane {n+1}' for n in range(self.U.shape[0]))
            yield ('Trial Masses', pd.DataFrame(tools.convert_cart_math(self.U),
                        index=_index, columns=['Mass']))

    def __repr__(self):
        '''
        Method to print out alpha value
        '''
        formatter = tools.InfoFormatter(name = 'Influence Coefficient Matrix', info_parameters=
                                        self._info())
        return ''.join(formatter.info())

    def save(self, file:str):
        '''
        Method to save influence coefficient values
        '''
        if isinstance(file, str):
            self.file = file
        np.save(file, self.value)


    def load(self, file:str):
        '''
        Method to load influence coefficient value
        '''
        if isinstance(file, str):
            self.file = file + '.npy'
        _matrix = np.load(self.file)
        self.add(direct_matrix=_matrix)

    @property
    def shape(self):
        '''
        returns shape of Influence coefficient matrix
        (no. Measuring Points, no. Balancing Planes)
        '''
        if (self.M is not None) and (self.N is not None):
            return (self.M, self.N)





class Condition():

    """
    Docstring for conditions.
    Condition is defined as speed or load or operating condition that is concerned in
    the balancing process.
    Conditions class is meant to be used for creating multispeed-multi_condition
    It is designed to arrange the conditions speeds and loads in explicit way.
    """

    def __init__(self:'condition', name:'string'=''):
        """
        Instantiate a conditions instance that will encapsulate all model speeds and loads
        name: optional name of Alpha
        """
        self.name = name

    def add(self, alpha:'Alpha instance', A:'initial_vibration numpy.array'):
        '''
        Method to add a new condition
        Args:
            alpha: Alpha class instance
            A: Initial vibration column array -> numpy array
        '''
        if isinstance(alpha, Alpha):
            self.alpha = alpha
        else:
            raise TypeError('alpha should be class Alpha.')
        try:
            _A_shape = A.shape
            # Test dimensions
            if A.ndim != 2:
                raise IndexError('A should be column vector of Mx1 dimension.')
            elif _A_shape[1] != 1:
                raise IndexError('A should be column vector of Mx1 dimension.')
            elif _A_shape[0] != self.alpha.value.shape[0]:
                raise IndexError('A and alpha should have the same 0 dimension(M).')
            else:
                self.A = A
        except AttributeError:
            raise TypeError('`A` should be passed as "numpy array"')


    def _info(self):
        '''
        Method to summarize the results for condition.
        '''
        if self.name:
            yield ('Name', self.name)
        if self.alpha is not None:
            yield ('Condition IC Matrix', str(self.alpha))
        if self.A is not None:
            _index = (f'Sensor {m+1}' for m in range(self.A.shape[0]))
            yield ('Initial Vibration', pd.DataFrame(tools.convert_cart_math(self.A),
                        index=_index, columns=['Vibration']))


    def __repr__(self):
        '''
        Method to print out condition value
        '''

        formatter = tools.InfoFormatter(name = 'Operation Condition', info_parameters=
                                        self._info(), level=2)

        return ''.join(formatter.info())

def test_save():
    alpha = Alpha(name='Model_IC')
    alpha.add(np.random.rand(3, 3))
    alpha.save('my_alpha')


if __name__ == '__main__':
    test_save()
    my_alpha = Alpha()
    my_alpha.load('my_alpha')
    print(my_alpha)






import time
from scipy.interpolate import make_interp_spline
import matplotlib.pyplot as plt
import scipy.optimize as opt
import numpy as np
import sys
import yaml
import pytest
import test_tools
import hsbalance as hs

def test_model_with_no_argument():
    with pytest.raises(TypeError) as e_info:
        model = hs.LeastSquares()
        assert 'Either (A and Alpha) or `conditions` should be assigned.' in str(e_info)
        A = np.random.rand(2,1)
        condition = hs.Conditions()
        model = hs.LeastSquares(A=A, conditions=[condition])
        assert 'Either (A and Alpha) or `conditions` should be assigned.' in str(e_info)
def test_faults():
    with pytest.raises(hs.CustomError) as error:
        hs.LeastSquares([1, 2], [1, 2])

@pytest.fixture()
def test_A():
    '''
    Creating alpha instance to test throwing faults
    '''
    A = np.random.uniform(0, 10, [2, 1])
    return A

@pytest.fixture()
def test_alpha():
    '''
    Creating alpha instance to test throwing faults
    '''
    alpha = hs.Alpha()
    value = np.random.uniform(0, 10, [2, 2])
    alpha.add(value + value * 1j)
    return alpha

def test_A_dim():
    '''
    Test the dimension of A to be 2x1
    '''
    real = np.random.uniform(0, 10, [1, 2])
    imag = np.random.uniform(0, 10, [1, 2])
    A = real + imag * 1j
    with pytest.raises(hs.CustomError) as e_info:
        hs.LeastSquares(test_A_dim, test_alpha)
    real = np.random.uniform(0, 10, [2, 2])
    imag = np.random.uniform(0, 10, [2, 2])
    A = real + imag * 1j
    with pytest.raises(hs.CustomError) as e_info:
        hs.LeastSquares(test_A_dim, test_alpha)


@pytest.fixture()
def n():
    '''
    tests for various matrix sizes
    for n = 1000 it took around 5 minutes on my machine
    '''
    return 250

@pytest.fixture()
def test_big_alpha(n):
    '''
    Creating alpha instance to test throwing faults
    '''
    alpha = hs.Alpha()
    real = np.random.uniform(0, 10, [n, n])
    imag = np.random.uniform(0, 10, [n, n])
    alpha.add(real + imag * 1j)
    return alpha

@pytest.fixture()
def test_big_A(n):
    '''
    Creating alpha instance to test throwing faults
    '''
    real = np.random.uniform(0, 10, [n, 1])
    imag = np.random.uniform(0, 10, [n, 1])
    return real + imag * 1j

@pytest.fixture
def test_model_LSQ(test_big_A, test_big_alpha):
    '''
    Creating a test model
    '''
    w =  hs.LeastSquares(test_big_A, test_big_alpha).solve()
    return w


@pytest.mark.timeout(100)
def test_big_matrix_LSQ(test_big_alpha, test_big_A, test_model_LSQ):
    '''
    Testing the perfromance of model
    '''
    start = time.time()
    error = hs.residual_vibration(test_big_alpha.value, test_model_LSQ, test_big_A)
    end = time.time()
    print(error.shape)
    print ('Time elapsed = ', end-start)
    np.testing.assert_allclose(error, 0, atol=1e-5)

# Testing Condition
def test_Condition(test_A, test_alpha):
    condition = hs.Condition()
    condition.add(test_alpha, test_A)
    assert condition.alpha == test_alpha
    np.testing.assert_allclose(condition.A, test_A)

    with pytest.raises(IndexError) as e_info:
        row_A = np.random.rand(5)
        condition.add(test_alpha, row_A)
    assert 'A should be column vector of Mx1 dimension' in str(e_info)

    with pytest.raises(IndexError) as e_info:
        square_A = np.random.rand(5, 5)
        condition.add(test_alpha, square_A)
    assert 'A should be column vector of Mx1 dimension' in str(e_info)

    with pytest.raises(IndexError) as e_info:
        mismatch_A = np.random.rand(3,1)
        condition.add(test_alpha, mismatch_A)
    assert 'A and alpha should have the same 0 dimension(M).' in str(e_info)

    with pytest.raises(TypeError) as e_info:
        A = [[3], [3], [3]]
        condition.add(test_alpha, A)
    assert 'numpy array' in str(e_info)

    with pytest.raises(IndexError) as e_info:
        alpha = hs.Alpha()
        alpha.add(np.random.rand(5,5))
        wrong_first_dim_A = np.random.rand(3, 1)
        condition.add(alpha, wrong_first_dim_A)
    print(condition.alpha.value, condition.A)
    assert 'same 0 dimension(M)' in str(e_info)

def test_model_conditions():
    test_alpha_1 = hs.Alpha()
    test_alpha_1.add(np.ones((2,2)))
    test_alpha_2 = hs.Alpha()
    test_alpha_2.add(np.ones((2,2)) * 2)
    test_A_1 = np.ones((2,1))
    test_A_2 = np.ones((2,1)) * 2
    condition_1 = hs.Condition()
    condition_1.add(test_alpha_1, test_A_1)
    condition_2 = hs.Condition()
    condition_2.add(test_alpha_2, test_A_2)
    model = hs.LeastSquares(conditions =[condition_1, condition_2])
    np.testing.assert_allclose(model.ALPHA , np.array([[1, 1],
                                                              [1, 1],
                                                              [2, 2],
                                                              [2, 2]]))
    np.testing.assert_allclose(model.A , np.array([[1],
                                                   [1],
                                                   [2],
                                                   [2]]))

def test_info_model():
    # Set random data
    np.random.seed(42)
    m = 3
    n = 3
    real = np.random.rand(m, n)
    imag = np.random.rand(m, n)
    comp= real + imag*1j
    alpha1 = hs.Alpha()
    alpha1.add(comp)
    real = np.random.rand(m, n)
    imag = np.random.rand(m, n)
    comp= real + imag*1j
    alpha2 = hs.Alpha()
    alpha2.add(comp)
    # Condition logging and printing
    condition1 = hs.Condition(name='Speed 1300')
    condition1.add(alpha2, A=np.random.rand(m,1))
    condition2 = hs.Condition(name='Speed 2500')
    condition2.add(alpha2, A=np.random.rand(m,1))
    model = hs.LeastSquares(conditions=[condition1, condition2])
    model.solve()
    angles = np.arange(100,300,10)  # angles
    split1 = model.create_split()
    split1.split_setup(0, max_number_weights_per_hole=1, holes_available=[angles]
                                                   ,weights_available=[0.1])
    split2 = model.create_split()
    split2.split_setup(1, max_number_weights_per_hole=1, holes_available=[angles]
                                                   ,weights_available=[0.2])
    split1.split_solve()
    split2.split_solve()
    split1.update(confirm=True)
    split2.update(confirm=True)
    print(model.info())


if __name__ == '__main__':
    def test_performance(n):
        alpha = hs.Alpha()
        real = np.random.uniform(0, 10, [n, n])
        imag = np.random.uniform(0, 10, [n, n])
        alpha.add(real + imag * 1j)
        real = np.random.uniform(0, 10, [n, 1])
        imag = np.random.uniform(0, 10, [n, 1])
        A= real + imag * 1j
        start = time.time()
        w =  hs.LeastSquares(A, alpha).solve()
        error = hs.residual_vibration(alpha.value, w, A)
        t = time.time() - start
        return round(t, 2)
    performance_time = []
    N = [2, 10, 50, 100, 200, 400, 600, 800]
    for n in N:
        performance_time.append(test_performance(n))
    print(N, performance_time)
    spline = make_interp_spline(N, performance_time)
    x = np.linspace(min(N), max(N), 500)
    y  = spline(x)
    plt.plot(x, y, label="Performace Test")
    plt.xlabel('N (dimension of a Squared Influence Coeffecient Matrix)')
    plt.ylabel('Time (seconds)')
    plt.title('Performance Test of LeastSquares model')
    plt.savefig('../data/performace_test.png')
    plt.show()
